#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include "ports.h"

int   physical_memory_fd = -1;
void *ports_base;

void open_ports(u_int32_t base_addr, u_int32_t span) {
    if (physical_memory_fd < 0)
        physical_memory_fd = open("/dev/mem", O_RDWR | O_SYNC);
    if (physical_memory_fd < 0) {
        printf("Failed to open /dev/mem for read/write -- permissions perhaps?\n");
        exit(-__LINE__);
    }
    ports_base = mmap(NULL, span, PROT_READ | PROT_WRITE, MAP_SHARED, physical_memory_fd, base_addr);
    if (ports_base == MAP_FAILED) {
        printf("Failed to mmap physical memory page.\n");
        exit(-__LINE__);
    }
}

void close_ports() {
    if (physical_memory_fd >= 0)
        close(physical_memory_fd);
}

#define FIFO_FILL_LEVEL     0
#define FIFO_STATUS         4
#define FIFO_STATUS_FULL    0x01
#define FIFO_STATUS_EMPTY   0x02
#define FIFO_EVENT          8
#define FIFO_INTERRUPT_ENABLE   12
#define FIFO_ALMOST_FULL    16
#define FIFO_ALMOST_EMPTY   20

struct  fifo32 {
    volatile u_int32_t  *data_port;
    volatile u_int32_t  *fill_level_port;
    volatile u_int32_t  *status_port;
};

void fifo32_init(
    struct fifo32      *fifo,
    u_int32_t   data_address,
    u_int32_t   status_address) {
    fifo->data_port = (u_int32_t *) (data_address + (u_int32_t)ports_base);
    fifo->fill_level_port = (u_int32_t *) (status_address + FIFO_FILL_LEVEL + (u_int32_t)ports_base);
    fifo->status_port = (u_int32_t *) (status_address + FIFO_STATUS + (u_int32_t)ports_base);

    printf("ports_base = %p\n", ports_base);
    printf("data_port = %p\n", fifo->data_port);
    printf("fill_level_port = %p\n", fifo->fill_level_port);
    printf("status_port = %p\n", fifo->status_port);
}

int fifo32_empty(struct fifo32 *fifo) {
    return (*fifo->status_port & FIFO_STATUS_EMPTY) ? 1 : 0;
}

int fifo32_full(struct fifo32 *fifo) {
    return (*fifo->status_port & FIFO_STATUS_FULL) ? 1 : 0;
}

void fifo32_write(struct fifo32 *fifo, u_int32_t  v) {
    while (fifo32_full(fifo))
        ;
    *fifo->data_port = v;
}

u_int32_t fifo32_read(struct fifo32 *fifo) {
    while (fifo32_empty(fifo))
        ;
    return *fifo->data_port;
}

u_int32_t fifo32_entries(struct fifo32 *fifo) {
    return *fifo->fill_level_port;
}

int main() {
    struct  fifo32  input, output;

    printf("Opening ports\n");fflush(stdout);
    open_ports(0xc0000000, 512);
    printf("Port range opened\n");fflush(stdout);

    fifo32_init(&output, FIFO_0_IN_BASE, FIFO_0_IN_CSR_BASE);
    fifo32_init(&input, FIFO_1_OUT_BASE, FIFO_1_OUT_CSR_BASE);

    // drain the input fifo of any left over cruft
    while (!fifo32_empty(&input))
        fifo32_read(&input);

    u_int32_t   i = 0, v;
    for (i = 0; i < 1024; i++) {

        fifo32_write(&output, i);
        if ((v = fifo32_read(&input)) != i) {
            printf("Error: wrote %d but read back %d\n", i, v);
            exit(-__LINE__);
        }
    }

    // a more complex example
    for (i = 0; i < 16; i++)
        fifo32_write(&output, i);

    for (i = 0; i < 16; i++)
        if ((v = fifo32_read(&input)) != i) {
            printf("Error: wrote %d but read back %d\n", i, v);
            exit(-__LINE__);
        }

    printf("Test pasted!\n");

    close_ports();
}


